@@ -2,7 +2,7 @@
 #
 #  Net::Server::Proto::UNIX - Net::Server Protocol module
 #
-#  Copyright (C) 2001-2017
+#  Copyright (C) 2001-2022
 #
 #    Paul Seamons <paul@seamons.com>
 #
@@ -19,7 +19,7 @@ package Net::Server::Proto::UNIX;
 
 use strict;
 use base qw(IO::Socket::UNIX);
-use Socket qw(SOCK_STREAM SOCK_DGRAM);
+use Socket qw(SOCK_STREAM SOCK_DGRAM SOMAXCONN);
 
 sub NS_proto { 'UNIX' }
 sub NS_port   { my $sock = shift; ${*$sock}{'NS_port'}   = shift if @_; return ${*$sock}{'NS_port'}   }
@@ -54,7 +54,7 @@ sub object {
     $sock->NS_port($port);
     $sock->NS_listen(defined($info->{'listen'}) ? $info->{'listen'}
                     : defined($server->{'server'}->{'listen'}) ? $server->{'server'}->{'listen'}
-                    : Socket::SOMAXCONN());
+                    : SOMAXCONN);
     return $sock;
 }
 
@@ -146,11 +146,11 @@ Protocol module for Net::Server.  This module implements the UNIX
 SOCK_STREAM socket type.  See L<Net::Server::Proto>.
 
 Any sockets created during startup will be chown'ed to the user and
-group specified in the starup arguments.
+group specified in the startup arguments.
 
 =head1 PARAMETERS
 
-The following paramaters may be specified in addition to normal
+The following parameters may be specified in addition to normal
 command line parameters for a Net::Server.  See L<Net::Server> for
 more information on reading arguments.
 
